

WARN = 3
ERROR = 7


class Feedback(object):

    def __init__(self):
        self.messages = []
        self.has_errors = False

    def error(self, msg):
        self.messages.append((ERROR, msg))
        self.has_errors = True

    def warn(self, msg):
        self.messages.append((WARN, msg))

    def __iter__(self):
        return iter(self.messages)

    def __nonzero__(self):
        return bool(self.messages)

    def get_errors(self):
        return [m for (m_kind, m) in self.messages if m_kind == ERROR]

    def get_warnings(self):
        return [m for (m_kind, m) in self.messages if m_kind == WARN]

    def extend(self, other):
        self.messages.extend(other.messages)
        if not self.has_errors and other.has_errors:
            self.has_errors = True



from copy import deepcopy
from contextlib import contextmanager

import errno
import logging
from logging.config import dictConfig as logConfig

import json
import os
from os.path import (
    abspath,
    expanduser,
    isabs,
    isdir,
    join as path_join,
    exists as path_exists,
)

import stat
import subprocess
import time
import tempfile
from urllib2 import (
    HTTPError,
    URLError,
    urlopen,
)
import zipfile

try:
    from yaml import CSafeLoader, CSafeDumper
    SafeLoader, SafeDumper = CSafeLoader, CSafeDumper
except ImportError:
    from yaml import SafeLoader

import yaml


class ErrorExit(Exception):

    def __init__(self, error=None):
        self.error = error


class DeploymentError(Exception):
    """One or more errors occurred during the deployment preparation."""

    def __init__(self, errors):
        self.errors = errors
        super(DeploymentError, self).__init__(errors)

    def __str__(self):
        return '\n'.join(self.errors)


STORE_URL = "https://store.juju.ubuntu.com"


# Utility functions
def yaml_dump(value):
    return yaml.dump(value, default_flow_style=False)


def yaml_load(value):
    return yaml.load(value, Loader=SafeLoader)


# We're not using safe dumper because we're using other custom
# representers as well.
def _unicode_representer(dumper, uni):
    node = yaml.ScalarNode(tag=u'tag:yaml.org,2002:str', value=uni)
    return node

yaml.add_representer(unicode, _unicode_representer)


DEFAULT_LOGGING = """
version: 1
formatters:
    standard:
        format: '%(asctime)s %(message)s'
        datefmt: "%Y-%m-%d %H:%M:%S"
    detailed:
        format: '%(asctime)s [%(levelname)s] %(name)s: %(message)s'
        datefmt: "%Y-%m-%d %H:%M:%S"
handlers:
    console:
        class: logging.StreamHandler
        formatter: standard
        level: DEBUG
        stream: ext://sys.stderr
loggers:
    deployer:
        level: INFO
        propogate: true
    deploy.cli:
        level: DEBUG
        propogate: true
    deploy.charm:
        level: DEBUG
        propogate: true
    deploy.env:
        level: DEBUG
        propogate: true
    deploy.deploy:
        level: DEBUG
        propogate: true
    deploy.importer:
        level: DEBUG
        propogate: true
    "":
        level: INFO
        handlers:
            - console
"""


def setup_logging(verbose=False, debug=False, stream=None):
    config = yaml_load(DEFAULT_LOGGING)
    log_options = {}
    if verbose:
        log_options.update({"loggers": {
            "deployer": {"level": "DEBUG", "propogate": True}}})
        #log_options.update({"loggers": {""
    if debug:
        log_options.update(
            {"handlers": {"console": {"formatter": "detailed"}}})
    config = dict_merge(config, log_options)
    logConfig(config)

    # Allow tests to reuse this func to mass configure log streams.
    if stream:
        root = logging.getLogger()
        previous = root.handlers[0]
        root.handlers[0] = current = logging.StreamHandler(stream)
        current.setFormatter(previous.formatter)
        return stream


@contextmanager
def temp_file():
    t = tempfile.NamedTemporaryFile()
    try:
        yield t
    finally:
        t.close()


def extract_zip(zip_path, dir_path):
    zf = zipfile.ZipFile(zip_path, "r")
    for info in zf.infolist():
        mode = info.external_attr >> 16
        if stat.S_ISLNK(mode):
            source = zf.read(info.filename)
            target = os.path.join(dir_path, info.filename)
            if os.path.exists(target):
                os.remove(target)
            os.symlink(source, target)
            continue
        extract_path = zf.extract(info, dir_path)
        os.chmod(extract_path, mode)


UNITS_DICT = {
    'M': 1,
    'G': 1024,
    'T': 1024 * 1024,
    'P': 1024 * 1024 * 1024,
}


def _to_number(value):
    """Convert a string to a numeric.

    The returned value is either an int or a float, depending on the input.
    Raises a ValueError if the value cannot be parsed.
    """
    try:
        num = int(value)
    except ValueError:
        num = float(value)
    return num


def _convert_units_specifier(value):
    """Convert a string that may have a unit specifier.

    Given a string possibly containing a unit specifier, return the
    the string representing the number and the units multiplier.
    """
    units = UNITS_DICT.get(value[-1], None)
    if units is not None:
        value = value[:-1]
    else:
        units = 1
    return value, units


def parse_constraints(value):
    """Parse the constraints, converting size specifiers into ints.

    Specifiers of 'M' and 'G' are supported.  The resulting value is the number
    of megabytes.

    The input is either a string of the form "k1=v1 ... kn=vn" or a dictionary.
    """
    if value is None:
        return value
    constraints_with_units = ['mem', 'root-disk']
    numerics = ['cpu-cores', 'cpu-power'] + constraints_with_units
    constraints = {}
    if isinstance(value, dict):
        constraints.update(value)
    else:
        pairs = value.strip().split()
        for item in pairs:
            k, v = item.split('=')
            constraints[k] = v

    for k, v in constraints.items():
        if k in numerics:
            units = 1
            if k in constraints_with_units:
                v, units = _convert_units_specifier(v)
            try:
                v = _to_number(v) * units
            except ValueError:
                raise ValueError(
                    'Constraint {} has invalid value {}'.format(
                        k, constraints[k]))
        constraints[k] = v
    return constraints


def _get_juju_home():
    jhome = os.environ.get("JUJU_HOME")
    if jhome is None:
        jhome = path_join(os.environ.get('HOME'), '.juju')
    return jhome


def _check_call(params, log, *args, **kw):
    max_retry = kw.get('max_retry', None)
    cur = kw.get('cur_try', 1)
    try:
        cwd = abspath(".")
        if 'cwd' in kw:
            cwd = kw['cwd']
        stderr = subprocess.STDOUT
        if 'stderr' in kw:
            stderr = kw['stderr']
        output = subprocess.check_output(
            params, cwd=cwd, stderr=stderr, env=os.environ)
    except subprocess.CalledProcessError, e:
        if 'ignoreerr' in kw:
            return
        #print "subprocess error"
        #print " ".join(params), "\ncwd: %s\n" % cwd, "\n ".join(
        # ["%s=%s" % (k, os.environ[k]) for k in os.environ
        #     if k.startswith("JUJU")])
        #print e.output
        log.error(*args)
        log.error("Command (%s) Output:\n\n %s", " ".join(params), e.output)
        if not max_retry or cur > max_retry:
            raise ErrorExit(e)
        kw['cur_try'] = cur + 1
        log.error("Retrying (%s of %s)" % (cur, max_retry))
        time.sleep(1)
        output = _check_call(params, log, args, **kw)
    return output


# Utils from deployer 1
def relations_combine(onto, source):
    target = deepcopy(onto)
    # Support list of relations targets
    if isinstance(onto, list) and isinstance(source, list):
        target.extend(source)
        return target
    for (key, value) in source.items():
        if key in target:
            if isinstance(target[key], dict) and isinstance(value, dict):
                target[key] = relations_combine(target[key], value)
            elif isinstance(target[key], list) and isinstance(value, list):
                target[key] = list(set(target[key] + value))
        else:
            target[key] = value
    return target


def dict_merge(onto, source):
    target = deepcopy(onto)
    for (key, value) in source.items():
        if key == 'relations' and key in target:
            target[key] = relations_combine(target[key], value)
        elif (key in target and isinstance(target[key], dict) and
                isinstance(value, dict)):
            target[key] = dict_merge(target[key], value)
        else:
            target[key] = value
    return target


def resolve_include(fname, include_dirs):
    if isabs(fname):
        return fname
    for path in include_dirs:
        full_path = path_join(path, fname)
        if path_exists(full_path):
            return full_path

    return None


def mkdir(path):
    """Create a leaf directory and all intermediate ones.

    Also expand ~ and ~user constructions.
    If path exists and it's a directory, return without errors.
    """
    path = expanduser(path)
    try:
        os.makedirs(path)
    except OSError as err:
        # Re-raise the error if the target path exists but it is not a dir.
        if (err.errno != errno.EEXIST) or (not isdir(path)):
            raise


def _is_qualified_charm_url(url):
    """Test an URL to see if it is revisioned."""
    parts = url.rsplit('-', 1)
    return len(parts) > 1 and parts[-1].isdigit()


def get_qualified_charm_url(url):
    """Given a charm URL, if not revisioned, return the latest revisioned URL.

    If the URL is already revisioned, return it.
    Otherwise ask the Charm store for the latest revision and return that URL.
    """
    if _is_qualified_charm_url(url):
        return url
    info_url = "%s/charm-info?charms=%s" % (STORE_URL, url)
    try:
        fh = urlopen(info_url)
    except (HTTPError, URLError) as e:
        errmsg = '{} ({})'.format(e, info_url)
        raise DeploymentError([errmsg])
    content = json.loads(fh.read())
    rev = content[url]['revision']
    return "%s-%d" % (url, rev)


def get_env_name(param_env_name):
    """Get the environment name.

    """
    if param_env_name:
        return param_env_name
    elif os.environ.get("JUJU_ENV"):
        return os.environ['JUJU_ENV']

    juju_home = _get_juju_home()
    env_ptr = os.path.join(juju_home, "current-environment")
    if os.path.exists(env_ptr):
        with open(env_ptr) as fh:
            return fh.read().strip()

    with open(os.path.join(juju_home, 'environments.yaml')) as fh:
        conf = yaml_load(fh.read())
        if not 'default' in conf:
            raise ValueError("No Environment specified")
        return conf['default']



from deployer.service import Service
from .base import Base
from ..utils import parse_constraints


class ConstraintsTest(Base):

    def test_constraints(self):
        data = {
            'branch': 'lp:precise/mysql',
            'constraints': "instance-type=m1.small",
        }
        s = Service('db', data)
        self.assertEquals(s.constraints, "instance-type=m1.small")
        data = {
            'branch': 'lp:precise/mysql',
            'constraints': "cpu-cores=4 mem=2048M root-disk=10G",
        }
        s = Service('db', data)
        c = parse_constraints(s.constraints)
        self.assertEquals(s.constraints, "cpu-cores=4 mem=2048M root-disk=10G")
        self.assertEqual(c['cpu-cores'], 4)
        self.assertEqual(c['mem'], 2048)
        self.assertEqual(c['root-disk'], 10 * 1024)

    def test_constraints_none(self):
        # If passed None, parse_constraints returns None.
        result = parse_constraints(None)
        self.assertIsNone(result)

    def test_constraints_dict(self):
        # The function can also accept a dict.
        value = {
            'mem': '1G',
            'cpu-cores': '5',
            'root-disk': '100',
        }
        result = parse_constraints(value)
        self.assertEqual(result['cpu-cores'], 5)
        self.assertEqual(result['mem'], 1024)
        self.assertEqual(result['root-disk'], 100)

    def test_constraints_accepts_no_spec(self):
        value = {
            'mem': '1',
            'root-disk': '10',
        }
        result = parse_constraints(value)
        self.assertEqual(result['mem'], 1)
        self.assertEqual(result['root-disk'], 10)

    def test_constraints_accept_M(self):
        value = {
            'mem': '1M',
            'root-disk': '10M',
        }
        mult = 1
        result = parse_constraints(value)
        self.assertEqual(result['mem'], 1 * mult)
        self.assertEqual(result['root-disk'], 10 * mult)

    def test_constraints_accept_G(self):
        value = {
            'mem': '1G',
            'root-disk': '10G',
        }
        result = parse_constraints(value)
        mult = 1024
        self.assertEqual(result['mem'], 1 * mult)
        self.assertEqual(result['root-disk'], 10 * mult)

    def test_constraints_accept_T(self):
        value = {
            'mem': '1T',
            'root-disk': '10T',
        }
        result = parse_constraints(value)
        mult = 1024 * 1024
        self.assertEqual(result['mem'], 1 * mult)
        self.assertEqual(result['root-disk'], 10 * mult)

    def test_constraints_accept_P(self):
        value = {
            'mem': '1P',
            'root-disk': '10P',
        }
        result = parse_constraints(value)
        mult = 1024 * 1024 * 1024
        self.assertEqual(result['mem'], 1 * mult)
        self.assertEqual(result['root-disk'], 10 * mult)

    def test_constraints_reject_other_sizes(self):
        value = {
            'mem': '1E',
        }
        with self.assertRaises(ValueError) as exc:
            result = parse_constraints(value)
        self.assertEqual('Constraint mem has invalid value 1E',
                         exc.exception.message)

    def test_other_numeric_constraints_have_no_units(self):
        # If any other numeric constraint gets a units specifier an error is
        # raised.
        keys = ['cpu-power', 'cpu-cores']
        for k in keys:
            value = {
                k: '1T',
            }
            with self.assertRaises(ValueError) as exc:
                parse_constraints(value)
            self.assertEqual('Constraint {} has invalid value 1T'.format(k),
                         exc.exception.message)

    def test_non_numerics_are_not_converted(self):
        # Constraints that expect strings are not affected by the parsing.
        keys = ['arch', 'container', 'tags']
        for k in keys:
            value = {
                k: '1T',
            }
            result = parse_constraints(value)
            self.assertEqual(result[k], '1T')



from feedback import Feedback


class Service(object):

    def __init__(self, name, svc_data):
        self.svc_data = svc_data
        self.name = name

    def __repr__(self):
        return "<Service %s>" % (self.name)

    @property
    def annotations(self):
        a = self.svc_data.get('annotations')
        if a is None:
            return a
        # core annotations only supports string key / values
        d = {}
        for k, v in a.items():
            d[str(k)] = str(v)
        return d

    @property
    def config(self):
        return self.svc_data.get('options', None)

    @property
    def constraints(self):
        return self.svc_data.get('constraints', None)

    @property
    def num_units(self):
        return int(self.svc_data.get('num_units', 1))

    @property
    def unit_placement(self):
        # Separate checks to support machine 0 placement.
        value = self.svc_data.get('to')
        if value is None:
            value = self.svc_data.get('force-machine')
        if value is not None and not isinstance(value, list):
            value = [value]
        return value and map(str, value) or []

    @property
    def expose(self):
        return self.svc_data.get('expose', False)


class ServiceUnitPlacement(object):

    def __init__(self, service, deployment, status):
        self.service = service
        self.deployment = deployment
        self.status = status

    @staticmethod
    def _format_placement(machine, container=None):
        if container:
            return "%s:%s" % (container, machine)
        else:
            return machine

    def validate(self):
        feedback = Feedback()

        unit_placement = self.service.unit_placement
        if unit_placement is None:
            return feedback

        if not isinstance(unit_placement, list):
            unit_placement = [unit_placement]
        unit_placement = map(str, unit_placement)

        services = dict([(s.name, s) for s in self.deployment.get_services()])

        for idx, p in enumerate(unit_placement):
            if ':' in p:
                container, p = p.split(':')
                if container not in ('lxc', 'kvm'):
                    feedback.error(
                        "Invalid service:%s placement: %s" % (
                            self.service.name, unit_placement[idx]))
            if '=' in p:
                p, u_idx = p.split("=")
                if not u_idx.isdigit():
                    feedback.error(
                        "Invalid service:%s placement: %s",
                        self.service.name, unit_placement[idx])
            if p.isdigit() and p == '0':
                continue
            elif p.isdigit():
                feedback.error(
                    "Service placement to machine not supported %s to %s",
                    self.service.name, unit_placement[idx])
            elif p in services:
                if services[p].unit_placement:
                    feedback.error(
                        "Nested placement not supported %s -> %s -> %s" % (
                            self.service.name, p, services[p].unit_placement))
            else:
                feedback.error(
                    "Invalid service placement %s to %s" % (
                        self.service.name, unit_placement[idx]))
        return feedback

    def get(self, unit_number):
        status = self.status
        svc = self.service

        unit_mapping = svc.unit_placement
        if not unit_mapping:
            return None
        if len(unit_mapping) <= unit_number:
            return None

        unit_placement = placement = str(unit_mapping[unit_number])
        container = None
        u_idx = unit_number

        if ':' in unit_placement:
            container, placement = unit_placement.split(":")
        if '=' in placement:
            placement, u_idx = placement.split("=")

        if placement.isdigit() and placement == "0":
            return self._format_placement(placement, container)

        with_service = status['services'].get(placement)
        if with_service is None:
            # Should be caught in validate relations but sanity check
            # for concurrency.
            self.deployment.log.error(
                "Service %s to be deployed with non existant service %s",
                svc.name, placement)
            # Prefer continuing deployment with a new machine rather
            # than an in-progress abort.
            return None

        svc_units = with_service['units']
        if len(svc_units) <= unit_number:
            self.deployment.log.warning(
                "Service:%s deploy-with Service:%s, but no with unit found",
                svc.name, placement)
            return None
        unit_names = svc_units.keys()
        unit_names.sort()
        machine = svc_units[unit_names[int(u_idx)]].get('machine')
        if not machine:
            self.deployment.log.warning(
                "Service:%s deploy-with unit missing machine %s",
                svc.name, unit_names[unit_number])
            return None
        return self._format_placement(machine, container)



import base64
import StringIO
import os
import unittest


from deployer.config import ConfigStack
from deployer.deployment import Deployment
from deployer.utils import setup_logging, ErrorExit

from .base import Base, TEST_OFFLINE


class DeploymentTest(Base):

    def setUp(self):
        self.output = setup_logging(
            debug=True, verbose=True, stream=StringIO.StringIO())

    def get_named_deployment(self, file_name, stack_name):
        return ConfigStack(
            [os.path.join(
                self.test_data_dir, file_name)]).get(stack_name)

    @unittest.skipIf(TEST_OFFLINE,
                     "Requires configured bzr launchpad id and network access")
    def test_deployer(self):
        d = ConfigStack(
            [os.path.join(
                self.test_data_dir, "blog.yaml")]).get('wordpress-prod')
        services = d.get_services()
        self.assertTrue([s for s in services if s.name == "newrelic"])

        # Ensure inheritance order reflects reality, instead of merge value.
        self.assertEqual(
            d.data['inherits'],
            ['wordpress-stage', 'wordpress-base', 'metrics-base'])

        # Fetch charms to verify late binding config values & validation.
        t = self.mkdir()
        os.mkdir(os.path.join(t, "precise"))
        d.repo_path = t
        d.fetch_charms()

        # Load up overrides and resolves
        d.load_overrides(["key=abc"])
        d.resolve()

        # Verify include-base64
        self.assertEqual(d.get_service('newrelic').config, {'key': 'abc'})
        self.assertEqual(
            base64.b64decode(d.get_service('blog').config['wp-content']),
            "HelloWorld")

        # TODO verify include-file

        # Verify relations
        self.assertEqual(
            list(d.get_relations()),
            [('blog', 'db'), ('blog', 'cache'), ('blog', 'haproxy')])

    def test_validate_placement_sorting(self):
        d = self.get_named_deployment("stack-placement.yaml", "stack")
        services = d.get_services()
        self.assertEqual(services[0].name, 'nova-compute')
        try:
            d.validate_placement()
        except ErrorExit:
            self.fail("Should not fail")

    def test_validate_invalid_placement_nested(self):
        d = self.get_named_deployment("stack-placement-invalid.yaml", "stack")
        services = d.get_services()
        self.assertEqual(services[0].name, 'nova-compute')
        try:
            d.validate_placement()
        except ErrorExit:
            pass
        else:
            self.fail("Should fail")

    def test_validate_invalid_placement_no_with_service(self):
        d = self.get_named_deployment(
            "stack-placement-invalid-2.yaml", "stack")
        services = d.get_services()
        self.assertEqual(services[0].name, 'nova-compute')
        try:
            d.validate_placement()
        except ErrorExit:
            pass
        else:
            self.fail("Should fail")

    def test_get_unit_placement(self):
        d = self.get_named_deployment("stack-placement.yaml", "stack")
        status = {
            'services': {
                'nova-compute': {
                    'units': {
                        'nova-compute/2': {'machine': '1'},
                        'nova-compute/3': {'machine': '2'},
                        'nova-compute/4': {'machine': '3'}}}}}
        placement = d.get_unit_placement('ceph', status)
        self.assertEqual(placement.get(0), '1')
        self.assertEqual(placement.get(1), '2')
        self.assertEqual(placement.get(2), None)

        placement = d.get_unit_placement('quantum', status)
        self.assertEqual(placement.get(0), 'lxc:1')
        self.assertEqual(placement.get(2), 'lxc:3')
        self.assertEqual(placement.get(3), None)

        placement = d.get_unit_placement('verity', status)
        self.assertEqual(placement.get(0), 'lxc:3')

        placement = d.get_unit_placement('mysql', status)
        self.assertEqual(placement.get(0), '0')

        placement = d.get_unit_placement('semper', status)
        self.assertEqual(placement.get(0), '3')

    def test_multiple_relations_no_weight(self):
        data = {"relations": {"wordpress": {"consumes": ["mysql"]},
                              "nginx": {"consumes": ["wordpress"]}}}
        d = Deployment("foo", data, include_dirs=())
        self.assertEqual(
            [('nginx', 'wordpress'), ('wordpress', 'mysql')],
            list(d.get_relations()))

    def test_multiple_relations_weighted(self):
        data = {
            "relations": {
                "keystone": {
                    "weight": 100,
                    "consumes": ["mysql"]
                },
                "nova-compute": {
                    "weight": 50,
                    "consumes": ["mysql"]
                },
                "glance": {
                    "weight": 70,
                    "consumes": ["mysql"]
                },
            }
        }
        d = Deployment("foo", data, include_dirs=())
        self.assertEqual(
            [('keystone', 'mysql'), ('glance', 'mysql'),
             ('nova-compute', 'mysql')], list(d.get_relations()))

    def test_getting_service_names(self):
        # It is possible to retrieve the service names.
        deployment = self.get_named_deployment("stack-placement.yaml", "stack")
        service_names = deployment.get_service_names()
        expected_service_names = [
            'ceph', 'mysql', 'nova-compute', 'quantum', 'semper', 'verity']
        self.assertEqual(set(expected_service_names), set(service_names))



from base64 import b64encode

import logging
import pprint
import os
import yaml

from .charm import Charm
from .feedback import Feedback
from .service import Service, ServiceUnitPlacement
from .relation import Endpoint
from .utils import path_join, yaml_dump, ErrorExit, resolve_include


class Deployment(object):

    log = logging.getLogger("deployer.deploy")

    def __init__(self, name, data, include_dirs, repo_path=""):
        self.name = name
        self.data = data
        self.include_dirs = include_dirs
        self.repo_path = repo_path

    @property
    def series(self):
        # Series could use a little help, charm series should be inferred
        # directly from a store url
        return self.data.get('series', 'precise')

    @property
    def series_path(self):
        return path_join(self.repo_path, self.series)

    def pretty_print(self):
        pprint.pprint(self.data)

    def get_service(self, name):
        if not name in self.data['services']:
            return
        return Service(name, self.data['services'][name])

    def get_services(self):
        services = []
        for name, svc_data in self.data.get('services', {}).items():
            services.append(Service(name, svc_data))
        services.sort(self._placement_sort)
        return services

    def get_service_names(self):
        """Return a sequence of service names for this deployment."""
        return self.data.get('services', {}).keys()

    @staticmethod
    def _placement_sort(svc_a, svc_b):
        if svc_a.unit_placement:
            if svc_b.unit_placement:
                return cmp(svc_a.name, svc_b.name)
            return 1
        if svc_b.unit_placement:
            return -1
        return cmp(svc_a.name, svc_b.name)

    def get_unit_placement(self, svc, status):
        if isinstance(svc, (str, unicode)):
            svc = self.get_service(svc)
        return ServiceUnitPlacement(svc, self, status)

    def get_relations(self):
        if 'relations' not in self.data:
            return

        # Strip duplicate rels
        seen = set()

        def check(a, b):
            k = tuple(sorted([a, b]))
            if k in seen:
                #self.log.warning(" Skipping duplicate relation %r" % (k,))
                return
            seen.add(k)
            return True

        # Support an ordered list of [endpoints]
        if isinstance(self.data['relations'], list):
            for end_a, end_b in self.data['relations']:
                # Allow shorthand of [end_a, [end_b, end_c]]
                if isinstance(end_b, list):
                    for eb in end_b:
                        if check(end_a, eb):
                            yield (end_a, eb)
                else:
                    if check(end_a, end_b):
                        yield (end_a, end_b)
            return

        # Legacy format (dictionary of dictionaries with weights)
        rels = {}
        for k, v in self.data['relations'].items():
            expanded = []
            for c in v['consumes']:
                expanded.append((k, c))
            by_weight = rels.setdefault(v.get('weight', 0), [])
            by_weight.extend(expanded)
        for k in sorted(rels, reverse=True):
            for r in rels[k]:
                if check(*r):
                    yield r
        #self.log.debug(
        #    "Found relations %s\n  %s" % (" ".join(map(str, seen))))

    def get_charms(self):
        for k, v in self.data.get('services', {}).items():
            yield Charm.from_service(k, self.series_path, v)

    def get_charm_for(self, svc_name):
        svc_data = self.data['services'][svc_name]
        return Charm.from_service(svc_name, self.series_path, svc_data)

    def fetch_charms(self, update=False, no_local_mods=False):
        if not os.path.exists(self.series_path):
            os.mkdir(self.series_path)
        for charm in self.get_charms():
            if charm.is_local():
                if charm.exists():
                    if no_local_mods:
                        if charm.is_modified():
                            self.log.warning(
                                "Charm %r has local modifications",
                                charm.path)
                            raise ErrorExit()
                    if update:
                        charm.update(build=True)
                    continue
            charm.fetch()

    def resolve(self, cli_overides=()):
        # Once we have charms definitions available, we can do verification
        # of config options.
        self.load_overrides(cli_overides)
        self.resolve_config()
        self.validate_relations()
        self.validate_placement()

    def load_overrides(self, cli_overrides=()):
        """Load overrides."""
        overrides = {}
        overrides.update(self.data.get('overrides', {}))

        for o in cli_overrides:
            key, value = o.split('=', 1)
            overrides[key] = value

        for k, v in overrides.iteritems():
            found = False
            for svc_name, svc_data in self.data['services'].items():
                charm = self.get_charm_for(svc_name)
                if k in charm.config:
                    if 'options' not in svc_data:
                        svc_data['options'] = {}
                    svc_data['options'][k] = v
                    found = True
            if not found:
                self.log.warning(
                    "Override %s does not match any charms", k)

    def resolve_config(self):
        """Load any lazy config values (includes), and verify config options.
        """
        self.log.debug("Resolving configuration")
        # XXX TODO, rename resolve, validate relations
        # against defined services
        feedback = Feedback()
        for svc_name, svc_data in self.data.get('services', {}).items():
            if not 'options' in svc_data:
                continue
            charm = self.get_charm_for(svc_name)
            config = charm.config
            options = {}

            for k, v in svc_data['options'].items():
                if not k in config:
                    feedback.error(
                        "Invalid config charm %s %s=%s" % (charm.name, k, v))
                    continue
                iv = self._resolve_include(svc_name, k, v)
                if isinstance(iv, Feedback):
                    feedback.extend(iv)
                    continue
                if iv is not None:
                    v = iv
                options[k] = v
            svc_data['options'] = options
        self._handle_feedback(feedback)

    def _resolve_include(self, svc_name, k, v):
        feedback = Feedback()
        for include_type in ["file", "base64"]:
            if (not isinstance(v, basestring)
                or not v.startswith(
                    "include-%s://" % include_type)):
                continue
            include, fname = v.split("://", 1)
            ip = resolve_include(fname, self.include_dirs)
            if ip is None:
                feedback.error(
                    "Invalid config %s.%s include not found %s" % (
                    svc_name, k, v))
                continue
            with open(ip) as fh:
                v = fh.read()
                if include_type == "base64":
                    v = b64encode(v)
                return v
        if feedback:
            return feedback

    def validate_relations(self):
        # Could extend to do interface matching against charms.
        services = dict([(s.name, s) for s in self.get_services()])
        feedback = Feedback()
        for e_a, e_b in self.get_relations():
            for ep in [Endpoint(e_a), Endpoint(e_b)]:
                if not ep.service in services:
                    feedback.error(
                        ("Invalid relation in config,"
                         " service %s not found, rel %s") % (
                             ep.service, "%s <-> %s" % (e_a, e_b)))
                    continue
        self._handle_feedback(feedback)

    def validate_placement(self):
        services = dict([(s.name, s) for s in self.get_services()])
        feedback = Feedback()
        for name, s in services.items():
            placement = self.get_unit_placement(s, {})
            feedback.extend(placement.validate())
        self._handle_feedback(feedback)

    def _handle_feedback(self, feedback):
        for e in feedback.get_errors():
            self.log.error(e)
        for w in feedback.get_warnings():
            self.log.warning(w)
        if feedback.has_errors:
            raise ErrorExit()

    def save(self, path):
        with open(path, "w") as fh:
            fh.write(yaml_dump(self.data))

    @staticmethod
    def to_yaml(dumper, deployment):
        return dumper.represent_dict(deployment.data)

yaml.add_representer(Deployment, Deployment.to_yaml)



from os.path import abspath, isabs, join, dirname

import logging
import os
import tempfile
import shutil
import urllib
import urlparse


from .deployment import Deployment
from .utils import ErrorExit, yaml_load, path_exists, dict_merge


class ConfigStack(object):

    log = logging.getLogger("deployer.config")

    def __init__(self, config_files, cli_series=None):
        self.config_files = config_files
        self.cli_series = cli_series
        self.data = {}
        self.yaml = {}
        self.include_dirs = []
        self.urlopen = urllib.urlopen
        self.load()

    def _yaml_load(self, config_file):
        if config_file in self.yaml:
            return self.yaml[config_file]

        if urlparse.urlparse(config_file).scheme:
            response = self.urlopen(config_file)
            if response.getcode() == 200:
                temp = tempfile.NamedTemporaryFile(delete=True)
                shutil.copyfileobj(response, temp)
                temp.flush()
                config_file = temp.name
            else:
                self.log.warning("Could not retrieve %s", config_file)
                raise ErrorExit()

        with open(config_file) as fh:
            try:
                self.yaml[config_file] = yaml_load(fh.read())
            except Exception, e:
                self.log.warning(
                    "Couldn't load config file @ %r, error: %s:%s",
                    config_file, type(e), e)
                raise

        return self.yaml[config_file]

    def keys(self):
        return sorted(self.data)

    def get(self, key):
        if not key in self.data:
            self.log.warning("Deployment %r not found. Available %s",
                             key, ", ".join(self.keys()))
            raise ErrorExit()
        deploy_data = self.data[key]
        deploy_data = self._resolve_inherited(deploy_data)
        if self.cli_series:
            deploy_data['series'] = self.cli_series
        return Deployment(
            key, deploy_data, self.include_dirs,
            repo_path=os.environ.get("JUJU_REPOSITORY", ""))

    def load(self):
        data = {}
        include_dirs = []
        for fp in self._resolve_included():
            if path_exists(fp):
                include_dirs.append(dirname(abspath(fp)))
            d = self._yaml_load(fp)
            data = dict_merge(data, d)
        self.data = data
        for k in ['include-config', 'include-configs']:
            if k in self.data:
                self.data.pop(k)
        self.include_dirs = include_dirs

    def _inherits(self, d):
        parents = d.get('inherits', ())
        if isinstance(parents, basestring):
            parents = [parents]
        return parents

    def _resolve_inherited(self, deploy_data):
        if not 'inherits' in deploy_data:
            return deploy_data
        inherits = parents = self._inherits(deploy_data)
        for parent_name in parents:
            parent = self.get(parent_name)
            inherits.extend(self._inherits(parent.data))
            deploy_data = dict_merge(parent.data, deploy_data)

        deploy_data['inherits'] = inherits
        return deploy_data

    def _includes(self, config_file):
        files = [config_file]
        d = self._yaml_load(config_file)

        incs = d.get('include-configs') or d.get('include-config')
        if isinstance(incs, basestring):
            inc_fs = [incs]
        else:
            inc_fs = incs

        if inc_fs:
            for inc_f in inc_fs:
                if not isabs(inc_f):
                    inc_f = join(dirname(config_file), inc_f)
                files.extend(self._includes(inc_f))

        return files

    def _resolve_included(self):
        files = []
        [files.extend(self._includes(cf)) for cf in self.config_files]
        return files
